@@ -3,6 +3,7 @@ from google.protobuf.json_format import MessageToDict
 from datetime import datetime, timezone, timedelta
 
 import pandas as pd
+import math 
 
 from finamgrpc.tradeapi.v1.auth.auth_service_pb2_grpc import AuthServiceStub
 from finamgrpc.tradeapi.v1.assets.assets_service_pb2_grpc import AssetsServiceStub
@@ -21,11 +22,46 @@ from finamgrpc.tradeapi.v1 import side_pb2
 from google.protobuf.timestamp_pb2 import Timestamp
 from google.type.interval_pb2 import Interval
 from google.type.decimal_pb2 import Decimal
-import asyncio
+from typing import Any
+from collections.abc import Callable, Iterable
 import threading
 import time 
 
 class FinamApi:
+    class StreamHandle:
+        """
+        Хендл на конкретный стрим (symbol + stream_key).
+
+        Пример использования:
+            trades = api.ensure_quote_stream(..., stream_key="trades")
+            snap = trades.get_snapshot()
+            trades.stop()
+        """
+
+        def __init__(self, api: FinamApi, symbol: str, stream_key: str):
+            self._api = api
+            self.symbol = symbol
+            self.stream_key = stream_key
+
+        def get_snapshot(self) -> dict | None:
+            """
+            Вернуть последний snapshot (ver + data) или None.
+            """
+            key = (self.symbol, self.stream_key)
+            with self._api._q_lock:
+                snap = self._api._q_snapshots.get(key)
+                return None if snap is None else {
+                    "ver": snap["ver"],
+                    "data": dict(snap["data"]),
+                }
+
+        def stop(self) -> None:
+            """Остановить именно этот стрим."""
+            self._api._stop_stream(self.symbol, self.stream_key)
+
+        def __repr__(self) -> str:
+            return f"<StreamHandle symbol={self.symbol!r} key={self.stream_key!r}>"
+
     """
     FinamApi provides methods to authenticate, fetch market data,
     account information, and manage orders via Finam's gRPC Trade API.
@@ -114,9 +150,11 @@ class FinamApi:
         }
 
         self._q_lock = threading.Lock()
-        self._q_snapshots = {}                      # {symbol: {"ver": int, "data": {...}}}
-        self._q_threads = {}                        # {symbol: Thread}
-        self._q_stops = {}  
+        # ключ = (symbol, stream_key)
+        self._q_threads: dict[tuple[str, str], threading.Thread] = {}
+        self._q_stops: dict[tuple[str, str], threading.Event] = {}
+        self._q_snapshots: dict[tuple[str, str], dict] = {}
+        self._q_handles: dict[tuple[str, str], FinamApi.StreamHandle] = {}
 
     def auth(self):
         """
@@ -520,7 +558,23 @@ class FinamApi:
         stub = MarketDataServiceStub(self.channel)
         req = marketdata_service_pb2.SubscribeLatestTradesRequest(symbol=symbol)
         return stub.SubscribeLatestTrades(req, metadata=(self.metadata,))
+
+
+    def stream_orderbook(self, symbol):
+        """
+        Open a streaming RPC for live trades on a symbol. Iterate over the returned generator to receive updates.
+
+        Args:
+            symbol (str)
+
+        Returns:
+            generator: Yields protobuf Trade messages
+        """
+        stub = MarketDataServiceStub(self.channel)
+        req = marketdata_service_pb2.SubscribeOrderBookRequest(symbol=symbol)
+        return stub.SubscribeOrderBook(req, metadata=(self.metadata,))
     
+
     def assets_params(self,symbol,account_id): 
         stub = AssetsServiceStub(self.channel)
         req = assets_service_pb2.GetAssetParamsRequest(symbol = symbol, account_id = account_id)
@@ -545,75 +599,241 @@ class FinamApi:
         res_dict = self.flatten_dict(res_dict)
         return res_dict
 
-    def ensure_quote_stream(self,symbol: str):
-        if symbol in self._q_threads and self._q_threads[symbol].is_alive():
-            return  # already running
+    def _to_float(self, x):
+        try:
+            return float(x)
+        except (TypeError, ValueError):
+            return math.nan
+
+
+    def _build_trade_snapshot(self, ev: Any) -> dict | None:
+        if not ev.trades:
+            return None
+        t = ev.trades[-1]
+        return {
+            "symbol": ev.symbol,
+            "trade_id": t.trade_id,
+            "side": t.side,
+            "price": self._to_float(t.price.value),
+            "size": self._to_float(t.size.value),
+            "ts": t.timestamp.seconds + t.timestamp.nanos / 1e9,
+        }
+
+    # ----------- пример snapshot_builder для стакана -----------
+
+    def _build_orderbook_snapshot(self, ev: Any) -> dict | None:
+        if not getattr(ev, "bids", None) and not getattr(ev, "offers", None):
+            return None
+
+        bids = [
+            {
+                "price": self._to_float(l.price.value),
+                "size": self._to_float(l.size.value),
+            }
+            for l in getattr(ev, "bids", [])
+        ]
+        offers = [
+            {
+                "price": self._to_float(l.price.value),
+                "size": self._to_float(l.size.value),
+            }
+            for l in getattr(ev, "offers", [])
+        ]
+
+        best_bid = bids[0]["price"] if bids else math.nan
+        best_ask = offers[0]["price"] if offers else math.nan
+
+        return {
+            "symbol": ev.symbol,
+            "bids": bids,
+            "offers": offers,
+            "best_bid": best_bid,
+            "best_ask": best_ask,
+            "ts": ev.timestamp.seconds + ev.timestamp.nanos / 1e9,
+        }
+
+    # ----------------- ОДНА ensure_quote_stream -----------------
+
+    def ensure_quote_stream(
+        self,
+        symbol: str,
+        stream_fn: Callable[[str], Iterable[Any]],
+        snapshot_builder: Callable[[Any], dict | None],
+        stream_key: str | None = None,
+    ) -> StreamHandle:
+        """
+        Универсальный запуск стрима.
+
+        symbol          — инструмент
+        stream_fn       — функция стрима (self.stream_trades / self.stream_orderbook / ...)
+        snapshot_builder— как из event собрать dict
+        stream_key      — идентификатор типа стрима ('trades', 'orderbook', ...).
+                          Если не задан, берём snapshot_builder.__name__.
+        Возвращает StreamHandle, через который читаем snapshot и останавливаем стрим.
+        """
+        if stream_key is None:
+            stream_key = snapshot_builder.__name__
+
+        key = (symbol, stream_key)
+
+        # если уже есть хендл и поток живой — просто возвращаем его
+        handle = self._q_handles.get(key)
+        if handle is not None:
+            th = self._q_threads.get(key)
+            if th is not None and th.is_alive():
+                return handle
+
+        # если хендла нет — создаём новый
+        handle = FinamApi.StreamHandle(self, symbol, stream_key)
+        self._q_handles[key] = handle
 
         stop_evt = threading.Event()
-        self._q_stops[symbol] = stop_evt
+        self._q_stops[key] = stop_evt
 
         def _runner():
             backoff = 1.0
             while not stop_evt.is_set():
                 try:
-                    for ev in self.stream_trades(symbol):
-                        if not ev.trades:
+                    for ev in stream_fn(symbol):
+                        if stop_evt.is_set():
+                            break
+
+                        snap = snapshot_builder(ev)
+                        if snap is None:
                             continue
-                        t = ev.trades[-1]  # last trade in the batch
-                        snap = {
-                            "symbol": ev.symbol,
-                            "trade_id": t.trade_id,
-                            "side": t.side,
-                            "price": float(t.price.value),
-                            "size": float(t.size.value),
-                            "ts": t.timestamp.seconds + t.timestamp.nanos / 1e9,
-                        }
+
                         with self._q_lock:
-                            rec = self._q_snapshots.get(symbol, {"ver": 0, "data": {}})
+                            rec = self._q_snapshots.get(key, {"ver": 0, "data": {}})
                             rec["ver"] += 1
                             rec["data"] = snap
-                            self._q_snapshots[symbol] = rec
-                        backoff = 1.0  # reset backoff after a good read
-                    # stream ended gracefully -> short sleep then reconnect
-                    time.sleep(1.0)
+                            self._q_snapshots[key] = rec
+
+                        backoff = 1.0  # успешное чтение – сброс backoff
+
+                    if not stop_evt.is_set():
+                        # стрим завершился “нормально” – пауза и переподключение
+                        time.sleep(1.0)
+
                 except grpc.RpcError as e:
-                    # handle UNAUTHENTICATED by refreshing your JWT (if you have a method for it)
                     if e.code() == grpc.StatusCode.UNAUTHENTICATED:
                         try:
-                            # self.refresh_jwt()   # implement if you have it
+                            # self.refresh_jwt()
                             pass
                         except Exception:
                             pass
                         backoff = 1.0
-                    # backoff for other errors
+
+                    if stop_evt.is_set():
+                        break
+
                     time.sleep(min(30.0, backoff))
                     backoff = min(30.0, backoff * 2)
 
-        th = threading.Thread(target=_runner, name=f"quote:{symbol}", daemon=True)
+        th = threading.Thread(
+            target=_runner, name=f"{stream_key}:{symbol}", daemon=True
+        )
         th.start()
-        self._q_threads[symbol] = th
+        self._q_threads[key] = th
 
-    def get_quote_snapshot(self,symbol: str) -> dict | None:
-        with self._q_lock:
-            snap = self._q_snapshots.get(symbol)
-            # returns like {"ver": 17, "data": {"symbol": "...", "bid": 99.1, "ask": 99.2, "last": 99.2, "ts": ...}}
-            return None if snap is None else {"ver": snap["ver"], "data": dict(snap["data"])}
+        return handle
 
-    def stop_quote_stream(self,symbol: str):
-        """
-        Gracefully stop the background quote stream for a given symbol.
-        """
-        evt = self._q_stops.get(symbol)
-        th = self._q_threads.get(symbol)
+    # ----------------- ВНУТРЕННИЙ STOP -----------------
+
+    def _stop_stream(self, symbol: str, stream_key: str) -> None:
+        key = (symbol, stream_key)
+        evt = self._q_stops.get(key)
+        th = self._q_threads.get(key)
 
         if evt:
-            evt.set()               # signal the runner loop to stop
+            evt.set()
         if th and th.is_alive():
-            th.join(timeout=2.0)    # wait a moment for it to finish
+            th.join(timeout=2.0)
 
-        # clean up dicts
-        self._q_stops.pop(symbol, None)
-        self._q_threads.pop(symbol, None)
-        self._q_snapshots.pop(symbol, None)
+        with self._q_lock:
+            self._q_stops.pop(key, None)
+            self._q_threads.pop(key, None)
+            self._q_snapshots.pop(key, None)
+            self._q_handles.pop(key, None)
 
+    # опционально: массовый стоп
+    def stop_all_streams(self) -> None:
+        for symbol, stream_key in list(self._q_threads.keys()):
+            self._stop_stream(symbol, stream_key)
 
+    def make_orderbook_snapshot_builder(self, symbol: str, depth: int = 20):
+        bids: dict[float, float] = {}
+        offers: dict[float, float] = {}
+        last_ts: float | None = None
+        debug_once = {"done": False}
+
+        def builder(ev):
+            nonlocal last_ts
+
+            # ---- найти rows ----
+            rows = getattr(ev, "rows", None)
+
+            if rows is None and getattr(ev, "order_book", None):
+                try:
+                    ob = ev.order_book[0]
+                    rows = ob.rows
+                except Exception:
+                    rows = None
+
+            if rows is None and isinstance(ev, (list, tuple)) and ev:
+                inner = ev[0]
+                rows = getattr(inner, "rows", None)
+                if rows is None and getattr(inner, "order_book", None):
+                    ob = inner.order_book[0]
+                    rows = ob.rows
+
+            if not rows:
+                if not debug_once["done"]:
+                    print("Orderbook builder: не могу найти rows в ev, type:", type(ev))
+                    debug_once["done"] = True
+                return None
+
+            # ---- применяем дельты ----
+            for row in rows:
+                price = float(row.price.value)
+
+                # определяем сторону и объем
+                if getattr(row, "buy_size", None) and row.buy_size.value:
+                    vol = float(row.buy_size.value)
+                    book = bids
+                else:
+                    vol = float(row.sell_size.value)
+                    book = offers
+
+                action = row.action  # это int: 0,1,2,3
+
+                # можем работать через константы:
+                # ADD/UPDATE → записать, REMOVE → удалить
+                if action in (row.ACTION_ADD, row.ACTION_UPDATE):
+                    if vol > 0:
+                        book[price] = vol
+                    else:
+                        book.pop(price, None)
+                elif action == row.ACTION_REMOVE:
+                    book.pop(price, None)
+                else:
+                    # ACTION_UNSPECIFIED или что-то новое – игнорим
+                    pass
+
+                ts = row.timestamp
+                last_ts = ts.seconds + ts.nanos / 1e9
+
+            # ---- собираем top-N уровней ----
+            best_bids = sorted(bids.items(), key=lambda x: x[0], reverse=True)[:depth]
+            best_offers = sorted(offers.items(), key=lambda x: x[0])[:depth]
+
+            snapshot = {
+                "symbol": symbol,
+                "ts": last_ts,
+                "bids": [{"price": p, "volume": v} for p, v in best_bids],
+                "offers": [{"price": p, "volume": v} for p, v in best_offers],
+                "best_bid": best_bids[0][0] if best_bids else math.nan,
+                "best_ask": best_offers[0][0] if best_offers else math.nan,
+            }
+            return snapshot
+
+        return builder
